from abc import abstractmethod
import json
from logging import Logger
from typing import Any, Dict, Optional

from aiohttp import ClientResponse, ClientSession, ClientTimeout
from aiohttp.client import DEFAULT_TIMEOUT
import jsonpickle

from .const import DEVICE_CLASS_SPEAKER, DEVICE_CLASS_TV

HTTP_OK = 200


class CommandBase(object):
    def __init__(self) -> None:
        self._url = ""

    @property
    def _method(self) -> str:
        return "PUT"

    @property
    def url(self) -> str:
        return self._url

    @url.setter
    def url(self, new_url: str) -> None:
        self._url = new_url

    def get_url(self) -> str:
        return self._url

    def get_method(self) -> str:
        return self._method

    @abstractmethod
    def process_response(self, json_obj: Dict[str, Any]) -> None:
        return None


class InfoCommandBase(CommandBase):
    def __init__(self) -> None:
        super(InfoCommandBase, self).__init__()

    @property
    def _method(self) -> str:
        return "GET"

    @property
    def url(self) -> str:
        return CommandBase.url.fget(self)

    @url.setter
    def url(self, new_url: str) -> None:
        CommandBase.url.fset(self, new_url)


class ProtoConstants(object):
    ACTION_MODIFY = "MODIFY"
    HEADER_AUTH = "AUTH"
    STATUS_SUCCESS = "success"
    RESPONSE_ITEM = "item"
    RESPONSE_ITEMS = "items"
    AUTH_TOKEN = "auth_token"
    CHALLENGE_TYPE = "challenge_type"
    PAIRING_REQ_TOKEN = "pairing_req_token"
    URI_NOT_FOUND = "URI_NOT_FOUND"

    class Item(object):
        HASHVAL = "hashval"
        CNAME = "cname"
        TYPE = "type"
        NAME = "name"
        VALUE = "value"
        METADATA = "metadata"
        ELEMENTS = "elements"


class KeyCodes(object):
    CODES = {
        DEVICE_CLASS_TV: {
            "SEEK_FWD": (2, 0),
            "SEEK_BACK": (2, 1),
            "PAUSE": (2, 2),
            "PLAY": (2, 3),
            "DOWN": (3, 0),
            "LEFT": (3, 1),
            "OK": (3, 2),
            "UP": (3, 3),
            "LEFT2": (3, 4),
            "RIGHT": (3, 5),
            "BACK": (4, 0),
            "SMARTCAST": (4, 3),
            "CC_TOGGLE": (4, 4),
            "INFO": (4, 6),
            "MENU": (4, 8),
            "HOME": (4, 15),
            "VOL_DOWN": (5, 0),
            "VOL_UP": (5, 1),
            "MUTE_OFF": (5, 2),
            "MUTE_ON": (5, 3),
            "MUTE_TOGGLE": (5, 4),
            "PIC_MODE": (6, 0),
            "PIC_SIZE": (6, 2),
            "INPUT_NEXT": (7, 1),
            "CH_DOWN": (8, 0),
            "CH_UP": (8, 1),
            "CH_PREV": (8, 2),
            "EXIT": (9, 0),
            "POW_OFF": (11, 0),
            "POW_ON": (11, 1),
            "POW_TOGGLE": (11, 2),
        },
        DEVICE_CLASS_SPEAKER: {
            "PAUSE": (2, 2),
            "PLAY": (2, 3),
            "VOL_DOWN": (5, 0),
            "VOL_UP": (5, 1),
            "MUTE_OFF": (5, 2),
            "MUTE_ON": (5, 3),
            "MUTE_TOGGLE": (5, 4),
            "POW_OFF": (11, 0),
            "POW_ON": (11, 1),
            "POW_TOGGLE": (11, 2),
        },
    }

    class KeyPressActions(object):
        KEY_DOWN = "KEYDOWN"
        KEY_UP = "KEYUP"
        KEY_PRESS = "KEYPRESS"


class Endpoints(object):
    ENDPOINTS = {
        DEVICE_CLASS_TV: {
            "BEGIN_PAIR": "/pairing/start",
            "FINISH_PAIR": "/pairing/pair",
            "CANCEL_PAIR": "/pairing/cancel",
            "INPUTS": "/menu_native/dynamic/tv_settings/devices/name_input",
            "CURR_INPUT": "/menu_native/dynamic/tv_settings/devices/current_input",
            "SET_INPUT": "/menu_native/dynamic/tv_settings/devices/current_input",
            "ESN": "/menu_native/dynamic/tv_settings/system/system_information/uli_information/esn",
            "POWER": "/state/device/power_mode",
            "KEY_PRESS": "/key_command/",
            "VOLUME": "/menu_native/dynamic/tv_settings/audio/volume",
        },
        DEVICE_CLASS_SPEAKER: {
            "BEGIN_PAIR": "/pairing/start",
            "FINISH_PAIR": "/pairing/pair",
            "CANCEL_PAIR": "/pairing/cancel",
            "INPUTS": "/menu_native/dynamic/audio_settings/input",
            "CURR_INPUT": "/menu_native/dynamic/audio_settings/input/current_input",
            "SET_INPUT": "/menu_native/dynamic/audio_settings/input/current_input",
            "ESN": "/menu_native/dynamic/audio_settings/system/system_information/uli_information/esn",
            "POWER": "/state/device/power_mode",
            "KEY_PRESS": "/key_command/",
            "VOLUME": "/menu_native/dynamic/audio_settings/audio/volume",
        },
    }


class CNames(object):
    class Audio(object):
        VOLUME = "volume"

    class ESN(object):
        ESN = "esn"


def get_json_obj(json_obj: Dict[str, Any], key: str) -> Any:
    key = key.lower()
    for k, v in json_obj.items():
        if k.lower() == key:
            return v
    return None


async def async_validate_response(web_response: ClientResponse) -> Dict[str, Any]:
    if HTTP_OK != web_response.status:
        raise Exception(
            "Device is unreachable? Status code: {0}".format(web_response.status)
        )
    try:
        data = json.loads(await web_response.text())
    except Exception:
        raise Exception("Failed to parse response: {0}".format(web_response.content))
    status_obj = get_json_obj(data, "status")
    if status_obj is None:
        raise Exception("Unknown response")
    result_status = get_json_obj(status_obj, "result")
    if result_status is None or result_status.lower() != ProtoConstants.STATUS_SUCCESS:
        raise Exception(
            "Unexpected response {0}: {1}".format(
                result_status, get_json_obj(status_obj, "detail")
            )
        )
    return data


async def async_invoke_api(
    ip: str,
    command: CommandBase,
    logger: Logger,
    timeout: int,
    headers: Optional[Dict[str, Any]] = None,
    log_api_exception: bool = True,
    session: Optional[ClientSession] = None,
) -> Any:
    if timeout:
        timeout = ClientTimeout(total=timeout)
    else:
        timeout = DEFAULT_TIMEOUT

    if headers is None:
        headers = {}

    try:
        method = command.get_method()
        if ":" not in ip:
            ip = ip + ":7345"
        url = "https://{0}{1}".format(ip, command.get_url())
        data = jsonpickle.encode(command, unpicklable=False)

        if session:
            if "get" == method.lower():
                response = await session.get(
                    url=url, headers=headers, ssl=False, timeout=timeout
                )
            else:
                headers["Content-Type"] = "application/json"
                response = await session.put(
                    url=url, data=str(data), headers=headers, ssl=False, timeout=timeout
                )

            json_obj = await async_validate_response(response)
        else:
            async with ClientSession() as local_session:
                if "get" == method.lower():
                    response = await local_session.get(
                        url=url, headers=headers, ssl=False, timeout=timeout
                    )
                else:
                    headers["Content-Type"] = "application/json"
                    response = await local_session.put(
                        url=url,
                        data=str(data),
                        headers=headers,
                        ssl=False,
                        timeout=timeout,
                    )

                json_obj = await async_validate_response(response)

        return command.process_response(json_obj)
    except Exception as e:
        if log_api_exception:
            logger.error("Failed to execute command: %s", e)
        return None


async def async_invoke_api_auth(
    ip: str,
    command: CommandBase,
    auth_token: str,
    logger: Logger,
    timeout: int,
    log_api_exception: bool = True,
    session: Optional[ClientSession] = None,
) -> Any:
    headers = {ProtoConstants.HEADER_AUTH: auth_token}
    return await async_invoke_api(
        ip,
        command,
        logger,
        timeout,
        headers,
        log_api_exception=log_api_exception,
        session=session,
    )



from typing import Any, Dict, List, Tuple

from .protocol import CommandBase, Endpoints, KeyCodes


class KeyPressEvent(object):
    def __init__(
        self,
        key_code: Tuple[int, int],
        action: str = KeyCodes.KeyPressActions.KEY_PRESS,
    ) -> None:
        self.CODESET = key_code[0]
        self.CODE = key_code[1]
        self.ACTION = action


class EmulateRemoteCommand(CommandBase):
    def __init__(self, key_codes: List[Tuple[int, int]], device_type: str) -> None:
        super(EmulateRemoteCommand, self).__init__()
        CommandBase.url.fset(self, Endpoints.ENDPOINTS[device_type]["KEY_PRESS"])
        # noinspection SpellCheckingInspection
        self.KEYLIST = []
        for key_code in key_codes:
            event = KeyPressEvent(key_code)
            self.KEYLIST.append(event)

    def process_response(self, json_obj: Dict[str, Any]) -> bool:
        return True



import time
from typing import Callable, List

from zeroconf import IPVersion, ServiceBrowser, ServiceInfo, Zeroconf


class ZeroconfDevice:
    def __init__(self, name: str, ip: str, port: int, model: str, id: str) -> None:
        self.name = name
        self.ip = ip
        self.port = port
        self.model = model
        self.id = id


class ZeroconfListener:
    def __init__(self, func: Callable[[ServiceInfo], None]) -> None:
        self._func = func

    def add_service(self, zeroconf: Zeroconf, type: str, name: str) -> None:
        self._func(zeroconf.get_service_info(type, name))


def discover(service_type: str, timeout: int = 10) -> List[ZeroconfDevice]:
    services = []

    def append_service(info: ServiceInfo) -> None:
        name = info.name[: -(len(info.type) + 1)]
        ip = info.parsed_addresses(IPVersion.V4Only)[0]
        port = info.port
        model = info.properties[b"name"].decode("utf-8")
        # id = info.properties[b"id"]
        # # handle id decode for various discovered use cases
        # if isinstance(id, bytes):
        #     try:
        #         int(id, 16)
        #     except Exception:
        #         id = id.hex()
        # else:
        #     id = None

        service = ZeroconfDevice(name, ip, port, model, None)
        services.append(service)

    zeroconf = Zeroconf()
    ServiceBrowser(zeroconf, service_type, ZeroconfListener(append_service))
    time.sleep(timeout)
    zeroconf.close()
    return services



from typing import Any, Dict, List, Optional

from .const import DEVICE_CLASS_SPEAKER
from .protocol import (
    CommandBase,
    Endpoints,
    InfoCommandBase,
    ProtoConstants,
    get_json_obj,
)


class VizioInput(object):
    def __init__(self, json_item: Dict[str, Any], is_extended_metadata: bool) -> None:
        self.id = int(get_json_obj(json_item, ProtoConstants.Item.HASHVAL))
        self.c_name = get_json_obj(json_item, ProtoConstants.Item.CNAME)
        self.type = get_json_obj(json_item, ProtoConstants.Item.TYPE)
        self.name = get_json_obj(json_item, ProtoConstants.Item.NAME)
        self.meta_name = None
        self.meta_data = None

        meta = get_json_obj(json_item, ProtoConstants.Item.VALUE)
        if meta is not None:
            if is_extended_metadata:
                self.meta_name = get_json_obj(meta, ProtoConstants.Item.NAME)
                self.meta_data = get_json_obj(meta, ProtoConstants.Item.METADATA)
            else:
                self.meta_name = meta

        if self.meta_name is None or "" == self.meta_name:
            self.meta_name = self.c_name


class GetInputsListCommand(InfoCommandBase):
    """Obtaining list of available inputs"""

    def __init__(self, device_type: str) -> None:
        super(GetInputsListCommand, self).__init__()
        InfoCommandBase.url.fset(self, Endpoints.ENDPOINTS[device_type]["INPUTS"])
        self._device_type = device_type

    def process_response(self, json_obj: Dict[str, Any]) -> Optional[List[VizioInput]]:
        items = get_json_obj(json_obj, ProtoConstants.RESPONSE_ITEMS)

        # Last input for sound bar is the current input so it needs to be removed before processing
        if self._device_type == DEVICE_CLASS_SPEAKER:
            items = items[:-1]

        inputs = []

        if items is not None:
            for itm in items:
                v_input = VizioInput(itm, True)
                inputs.append(v_input)

        return inputs


class GetCurrentInputCommand(InfoCommandBase):
    """Obtaining current input"""

    def __init__(self, device_type: str) -> None:
        super(GetCurrentInputCommand, self).__init__()
        InfoCommandBase.url.fset(self, Endpoints.ENDPOINTS[device_type]["CURR_INPUT"])

    def process_response(self, json_obj: Dict[str, Any]) -> Optional[VizioInput]:
        items = get_json_obj(json_obj, ProtoConstants.RESPONSE_ITEMS)
        v_input = None
        if len(items) > 0:
            v_input = VizioInput(items[0], False)
        return v_input


class ChangeInputCommand(CommandBase):
    def __init__(self, id_, name: str, device_type: str) -> None:
        super(ChangeInputCommand, self).__init__()
        CommandBase.url.fset(self, Endpoints.ENDPOINTS[device_type]["SET_INPUT"])
        self.VALUE = str(name)
        # noinspection SpellCheckingInspection
        self.HASHVAL = int(id_)
        self.REQUEST = ProtoConstants.ACTION_MODIFY

    def process_response(self, json_obj: Dict[str, Any]) -> bool:
        return True



from typing import Any, Dict, List, Optional

from .protocol import CNames, Endpoints, InfoCommandBase, ProtoConstants, get_json_obj


class SettingsItem(object):
    def __init__(self, json_obj: Dict[str, Any]) -> None:
        self.id = int(get_json_obj(json_obj, ProtoConstants.Item.HASHVAL))
        self.c_name = get_json_obj(json_obj, ProtoConstants.Item.CNAME)
        self.type = get_json_obj(json_obj, ProtoConstants.Item.TYPE)
        self.name = get_json_obj(json_obj, ProtoConstants.Item.NAME)
        self.value = get_json_obj(json_obj, ProtoConstants.Item.VALUE)
        self.options = []
        options = get_json_obj(json_obj, ProtoConstants.Item.ELEMENTS)
        if options is not None:
            for opt in options:
                self.options.append(opt)


class GetCurrentAudioCommand(InfoCommandBase):
    def __init__(self, device_type: str) -> None:
        super(GetCurrentAudioCommand, self).__init__()
        InfoCommandBase.url.fset(self, Endpoints.ENDPOINTS[device_type]["VOLUME"])

    @staticmethod
    def _get_items(json_obj: Dict[str, Any]) -> List[SettingsItem]:
        items = get_json_obj(json_obj, ProtoConstants.RESPONSE_ITEMS)
        if items is None:
            return []

        results = []
        for itm in items:
            item = SettingsItem(itm)
            results.append(item)

        return results

    def process_response(self, json_obj: Dict[str, Any]) -> int:
        items = self._get_items(json_obj)
        for itm in items:
            if itm.c_name.lower() == CNames.Audio.VOLUME:
                if itm.value is not None:
                    return int(itm.value)
        return 0


class GetESNCommand(InfoCommandBase):
    def __init__(self, device_type: str) -> None:
        super(GetESNCommand, self).__init__()
        InfoCommandBase.url.fset(self, Endpoints.ENDPOINTS[device_type]["ESN"])

    @staticmethod
    def _get_items(json_obj: Dict[str, Any]) -> List[SettingsItem]:
        items = get_json_obj(json_obj, ProtoConstants.RESPONSE_ITEMS)
        if items is None:
            return []

        results = []
        for itm in items:
            item = SettingsItem(itm)
            results.append(item)

        return results

    def process_response(self, json_obj: Dict[str, Any]) -> Optional[str]:
        items = self._get_items(json_obj)
        for itm in items:
            if itm.c_name.lower() == CNames.ESN.ESN:
                if itm.value is not None:
                    return itm.value

        return None



import asyncio
import logging
from typing import Any, List, Optional, Union

from aiohttp import ClientSession

from .cmd_input import (
    ChangeInputCommand,
    GetCurrentInputCommand,
    GetInputsListCommand,
    VizioInput,
)
from .cmd_pair import (
    BeginPairCommand,
    BeginPairResponse,
    CancelPairCommand,
    PairChallengeCommand,
    PairChallengeResponse,
)
from .cmd_power import GetPowerStateCommand
from .cmd_remote import EmulateRemoteCommand
from .cmd_settings import GetCurrentAudioCommand, GetESNCommand
from .const import (
    DEFAULT_DEVICE_CLASS,
    DEFAULT_TIMEOUT,
    DEVICE_CLASS_SOUNDBAR,
    DEVICE_CLASS_SPEAKER,
    DEVICE_CLASS_TV,
)
from .discovery import ZeroconfDevice, discover
from .protocol import CommandBase, KeyCodes, async_invoke_api, async_invoke_api_auth

_LOGGER = logging.getLogger(__name__)

MAX_VOLUME = {DEVICE_CLASS_TV: 100, DEVICE_CLASS_SPEAKER: 31}


class DeviceDescription(object):
    def __init__(self, ip: str, name: str, model: str, id: str) -> None:
        self.ip = ip
        self.name = name
        self.model = model
        self.id = id


class VizioAsync(object):
    def __init__(
        self,
        device_id: str,
        ip: str,
        name: str,
        auth_token: str = "",
        device_type: str = DEFAULT_DEVICE_CLASS,
        session: Optional[ClientSession] = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> None:
        self._device_type = device_type.lower()
        if self._device_type == DEVICE_CLASS_SOUNDBAR:
            _LOGGER.warning(
                f"The '{DEVICE_CLASS_SOUNDBAR}' device type has been deprecated and will be removed soon. Please use the '{DEVICE_CLASS_SPEAKER}' device type going forward"
            )
            self._device_type = DEVICE_CLASS_SPEAKER

        if (
            self._device_type != DEVICE_CLASS_TV
            and self._device_type != DEVICE_CLASS_SPEAKER
        ):
            raise Exception(
                f"Invalid device type specified. Use either '{DEVICE_CLASS_TV}' or '{DEVICE_CLASS_SPEAKER}'"
            )

        self._ip = ip
        self._name = name
        self._device_id = device_id
        self._auth_token = auth_token
        self._session = session
        self._timeout = timeout

    async def __invoke_api(
        self,
        cmd: CommandBase,
        log_api_exception: bool = True,
        custom_timeout: bool = True,
    ) -> Any:
        timeout = None
        if custom_timeout:
            timeout = self._timeout

        return await async_invoke_api(
            self._ip,
            cmd,
            _LOGGER,
            timeout,
            log_api_exception=log_api_exception,
            session=self._session,
        )

    async def __invoke_api_auth(
        self,
        cmd: CommandBase,
        log_api_exception: bool = True,
        custom_timeout: bool = True,
    ) -> Any:
        timeout = None
        if custom_timeout:
            timeout = self._timeout

        return await async_invoke_api_auth(
            self._ip,
            cmd,
            self._auth_token,
            _LOGGER,
            timeout,
            log_api_exception=log_api_exception,
            session=self._session,
        )

    async def __invoke_api_may_need_auth(
        self,
        cmd: CommandBase,
        log_api_exception: bool = True,
        custom_timeout: bool = True,
    ) -> Any:
        if self._auth_token is None or "" == self._auth_token:
            if self._device_type == DEVICE_CLASS_SPEAKER:
                return await self.__invoke_api(
                    cmd,
                    log_api_exception=log_api_exception,
                    custom_timeout=custom_timeout,
                )
            else:
                raise Exception(
                    f"Empty auth token. To target a speaker and bypass auth requirements, pass '{DEVICE_CLASS_SPEAKER}' as device_type"
                )
        return await self.__invoke_api_auth(
            cmd, log_api_exception=log_api_exception, custom_timeout=custom_timeout
        )

    async def __remote(
        self, key_list: Union[List[str], str], log_api_exception: bool = True
    ) -> bool:
        key_codes = []
        if isinstance(key_list, list) is False:
            key_list = [key_list]

        for key in key_list:
            if key not in KeyCodes.CODES[self._device_type]:
                _LOGGER.error(
                    "Key Code of '%s' not found for device type of '%s'",
                    key,
                    self._device_type,
                )
                return False
            else:
                key_codes.append(KeyCodes.CODES[self._device_type][key])

        result = await self.__invoke_api_may_need_auth(
            EmulateRemoteCommand(key_codes, self._device_type),
            log_api_exception=log_api_exception,
            custom_timeout=False,
        )
        return result is not None

    async def __remote_multiple(
        self, key_code: str, num: int, log_api_exception: bool = True
    ) -> bool:
        key_codes = []
        for ii in range(0, num):
            key_codes.append(key_code)
        return await self.__remote(key_codes, log_api_exception=log_api_exception)

    @staticmethod
    def discovery(timeout: int = 10) -> List[ZeroconfDevice]:
        return discover("_viziocast._tcp.local.", timeout)

    @staticmethod
    async def validate_ha_config(
        ip: str, auth_token: str, device_type: str, timeout: int = DEFAULT_TIMEOUT
    ) -> bool:
        return await VizioAsync(
            "", ip, "", auth_token, device_type, timeout=timeout
        ).can_connect()

    @staticmethod
    async def get_unique_id(
        ip: str,
        auth_token: str,
        device_type: str,
        timeout: int = DEFAULT_TIMEOUT,
        log_api_exception: bool = True,
    ) -> Optional[str]:
        return await VizioAsync(
            "", ip, "", auth_token, device_type, timeout=timeout
        ).get_esn(log_api_exception=log_api_exception)

    async def can_connect(self) -> bool:
        if (
            await self.__invoke_api_may_need_auth(
                GetPowerStateCommand(self._device_type), log_api_exception=False
            )
            is not None
        ):
            return True
        else:
            return False

    async def get_esn(self, log_api_exception: bool = True) -> Optional[str]:
        return await self.__invoke_api_may_need_auth(
            GetESNCommand(self._device_type), log_api_exception=log_api_exception
        )

    async def start_pair(
        self, log_api_exception: bool = True
    ) -> Optional[BeginPairResponse]:
        return await self.__invoke_api(
            BeginPairCommand(self._device_id, self._name, self._device_type),
            log_api_exception=log_api_exception,
        )

    async def stop_pair(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__invoke_api(
            CancelPairCommand(self._device_id, self._name, self._device_type),
            log_api_exception=log_api_exception,
        )

    async def pair(
        self, ch_type: int, token: int, pin: str, log_api_exception: bool = True
    ) -> Optional[PairChallengeResponse]:
        return await self.__invoke_api(
            PairChallengeCommand(
                self._device_id, ch_type, token, pin, self._device_type
            ),
            log_api_exception=log_api_exception,
        )

    async def get_inputs(
        self, log_api_exception: bool = True
    ) -> Optional[List[VizioInput]]:
        return await self.__invoke_api_may_need_auth(
            GetInputsListCommand(self._device_type), log_api_exception=log_api_exception
        )

    async def get_current_input(
        self, log_api_exception: bool = True
    ) -> Optional[VizioInput]:
        return await self.__invoke_api_may_need_auth(
            GetCurrentInputCommand(self._device_type),
            log_api_exception=log_api_exception,
        )

    async def get_power_state(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__invoke_api_may_need_auth(
            GetPowerStateCommand(self._device_type), log_api_exception=log_api_exception
        )

    async def pow_on(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote("POW_ON", log_api_exception=log_api_exception)

    async def pow_off(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote("POW_OFF", log_api_exception=log_api_exception)

    async def pow_toggle(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote("POW_TOGGLE", log_api_exception=log_api_exception)

    async def vol_up(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        return await self.__remote_multiple(
            "VOL_UP", num, log_api_exception=log_api_exception
        )

    async def vol_down(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        return await self.__remote_multiple(
            "VOL_DOWN", num, log_api_exception=log_api_exception
        )

    async def get_current_volume(self, log_api_exception: bool = True) -> Optional[int]:
        return await self.__invoke_api_may_need_auth(
            GetCurrentAudioCommand(self._device_type),
            log_api_exception=log_api_exception,
        )

    def get_max_volume(self) -> int:
        return MAX_VOLUME[self._device_type]

    async def ch_up(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        return await self.__remote_multiple(
            "CH_UP", num, log_api_exception=log_api_exception
        )

    async def ch_down(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        return await self.__remote_multiple(
            "CH_DOWN", num, log_api_exception=log_api_exception
        )

    async def ch_prev(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote("CH_PREV", log_api_exception=log_api_exception)

    async def mute_on(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote("MUTE_ON", log_api_exception=log_api_exception)

    async def mute_off(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote("MUTE_OFF", log_api_exception=log_api_exception)

    async def mute_toggle(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote("MUTE_TOGGLE", log_api_exception=log_api_exception)

    async def input_next(self, log_api_exception: bool = True) -> Optional[bool]:
        # HACK: Single call just invoking overlay menu with current input
        return await self.__remote_multiple(
            "INPUT_NEXT", 2, log_api_exception=log_api_exception
        )

    async def input_switch(
        self, name: str, log_api_exception: bool = True
    ) -> Optional[bool]:
        if isinstance(self, Vizio):
            cur_input = await super(Vizio, self).get_current_input(
                log_api_exception=log_api_exception
            )
        else:
            cur_input = await self.get_current_input(
                log_api_exception=log_api_exception
            )
        if cur_input is None:
            _LOGGER.error("Couldn't detect current input")
            return False
        return await self.__invoke_api_may_need_auth(
            ChangeInputCommand(cur_input.id, name, self._device_type),
            log_api_exception=log_api_exception,
            custom_timeout=False,
        )

    async def play(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote("PLAY", log_api_exception=log_api_exception)

    async def pause(self, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote("PAUSE", log_api_exception=log_api_exception)

    async def remote(self, key: str, log_api_exception: bool = True) -> Optional[bool]:
        return await self.__remote(key, log_api_exception=log_api_exception)

    def get_device_keys(self) -> List[str]:
        return KeyCodes.CODES[self._device_type].keys()


async def async_guess_device_type(
    ip: str, port: Optional[str] = None, timeout: int = DEFAULT_TIMEOUT
) -> str:
    """
    Attempts to guess the device type by getting power state with no auth
    token.

    NOTE:
    The `ip` and `port` values passed in have to be valid for the device in
    order for this to work. This function is being used as part of a zeroconf
    discovery workflow in HomeAssistant which is why it is safe to assume that
    `ip` and `port` are valid.
    """

    if port:
        if ":" in ip:
            raise Exception("Port can't be included in both `ip` and `port` parameters")

        device = VizioAsync(
            "test", ip + ":" + port, "test", "", DEVICE_CLASS_SPEAKER, timeout=timeout
        )
    else:
        if ":" not in ip:
            _LOGGER.warning(
                "May not return correct device type since a port was not specified."
            )
        device = VizioAsync(
            "test", ip, "test", "", DEVICE_CLASS_SPEAKER, timeout=timeout
        )

    if await device.can_connect():
        return DEVICE_CLASS_SPEAKER
    else:
        return DEVICE_CLASS_TV


class Vizio(VizioAsync):
    def __init__(
        self,
        device_id: str,
        ip: str,
        name: str,
        auth_token: str = "",
        device_type: str = DEFAULT_DEVICE_CLASS,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> None:
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        super(Vizio, self).__init__(
            device_id, ip, name, auth_token, device_type, session=None, timeout=timeout
        )

    def __del__(self) -> None:
        self.loop.close()

    @staticmethod
    def discovery(timeout: int = 10) -> List[ZeroconfDevice]:
        return VizioAsync.discovery(timeout)

    @staticmethod
    def validate_ha_config(
        ip: str, auth_token: str, device_type: str, timeout: int = DEFAULT_TIMEOUT
    ) -> bool:
        return Vizio("", ip, "", auth_token, device_type, timeout).can_connect()

    @staticmethod
    def get_unique_id(
        ip: str,
        auth_token: str,
        device_type: str,
        timeout: int = DEFAULT_TIMEOUT,
        log_api_exception: bool = True,
    ) -> Optional[str]:
        return Vizio("", ip, "", auth_token, device_type, timeout).get_esn(
            log_api_exception=log_api_exception
        )

    def can_connect(self) -> bool:
        return self.loop.run_until_complete(
            self.loop.create_task(super(Vizio, self).can_connect())
        )

    def get_esn(self, log_api_exception: bool = True) -> Optional[str]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).get_esn(log_api_exception=log_api_exception)
            )
        )

    def start_pair(self, log_api_exception: bool = True) -> Optional[BeginPairResponse]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).start_pair(log_api_exception=log_api_exception)
            )
        )

    def stop_pair(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).stop_pair(log_api_exception=log_api_exception)
            )
        )

    def pair(
        self, ch_type: str, token: str, pin: str, log_api_exception: bool = True
    ) -> Optional[PairChallengeResponse]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).pair(
                    ch_type, token, pin, log_api_exception=log_api_exception
                )
            )
        )

    def get_inputs(self, log_api_exception: bool = True) -> Optional[List[VizioInput]]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).get_inputs(log_api_exception=log_api_exception)
            )
        )

    def get_current_input(self, log_api_exception: bool = True) -> Optional[VizioInput]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).get_current_input(
                    log_api_exception=log_api_exception
                )
            )
        )

    def get_power_state(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).get_power_state(log_api_exception=log_api_exception)
            )
        )

    def pow_on(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).pow_on(log_api_exception=log_api_exception)
            )
        )

    def pow_off(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).pow_off(log_api_exception=log_api_exception)
            )
        )

    def pow_toggle(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).pow_toggle(log_api_exception=log_api_exception)
            )
        )

    def vol_up(self, num: int = 1, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).vol_up(num, log_api_exception=log_api_exception)
            )
        )

    def vol_down(self, num: int = 1, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).vol_down(num, log_api_exception=log_api_exception)
            )
        )

    def get_current_volume(self, log_api_exception: bool = True) -> Optional[int]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).get_current_volume(
                    log_api_exception=log_api_exception
                )
            )
        )

    def get_max_volume(self) -> int:
        return super(Vizio, self).get_max_volume()

    def ch_up(self, num: int = 1, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).ch_up(num, log_api_exception=log_api_exception)
            )
        )

    def ch_down(self, num: int = 1, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).ch_down(num, log_api_exception=log_api_exception)
            )
        )

    def ch_prev(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).ch_prev(log_api_exception=log_api_exception)
            )
        )

    def mute_on(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).mute_on(log_api_exception=log_api_exception)
            )
        )

    def mute_off(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).mute_off(log_api_exception=log_api_exception)
            )
        )

    def mute_toggle(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).mute_toggle(log_api_exception=log_api_exception)
            )
        )

    def input_next(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).input_next(log_api_exception=log_api_exception)
            )
        )

    def input_switch(self, name: str, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).input_switch(
                    name, log_api_exception=log_api_exception
                )
            )
        )

    def play(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).play(log_api_exception=log_api_exception)
            )
        )

    def pause(self, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).pause(log_api_exception=log_api_exception)
            )
        )

    def remote(self, key: str, log_api_exception: bool = True) -> Optional[bool]:
        return self.loop.run_until_complete(
            self.loop.create_task(
                super(Vizio, self).remote(key, log_api_exception=log_api_exception)
            )
        )

    def get_device_keys(self) -> List[str]:
        return super(Vizio, self).get_device_keys()


def guess_device_type(
    ip: str, port: Optional[str] = None, timeout: int = DEFAULT_TIMEOUT
) -> str:
    """
    Attempts to guess the device type by getting power state with no auth
    token.

    NOTE:
    The `ip` and `port` values passed in have to be valid for the device in
    order for this to work. This function is being used as part of a zeroconf
    discovery workflow in HomeAssistant which is why it is safe to assume that
    `ip` and `port` are valid.
    """

    try:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop.run_until_complete(
            loop.create_task(async_guess_device_type(ip, port, timeout))
        )
    finally:
        loop.close()



from typing import Any, Dict

from .protocol import Endpoints, InfoCommandBase, ProtoConstants, get_json_obj


class GetPowerStateCommand(InfoCommandBase):
    def __init__(self, device_type: str) -> None:
        super(GetPowerStateCommand, self).__init__()
        InfoCommandBase.url.fset(self, Endpoints.ENDPOINTS[device_type]["POWER"])

    def process_response(self, json_obj: Dict[str, Any]) -> bool:
        items = get_json_obj(json_obj, ProtoConstants.RESPONSE_ITEMS)
        is_awake = False
        if len(items) > 0:
            is_awake = 1 == get_json_obj(items[0], ProtoConstants.Item.VALUE)

        return is_awake



from typing import Any, Dict

from .protocol import CommandBase, Endpoints, ProtoConstants, get_json_obj


class PairCommandBase(CommandBase):
    def __init__(self, device_id: str, device_type: str, endpoint: str) -> None:
        super(PairCommandBase, self).__init__()
        CommandBase.url.fset(self, Endpoints.ENDPOINTS[device_type][endpoint])
        self.DEVICE_ID = device_id


class BeginPairResponse(object):
    def __init__(self, ch_type: str, token: str) -> None:
        self.ch_type = ch_type
        self.token = token


class BeginPairCommand(PairCommandBase):
    """Initiating pairing process."""

    def __init__(self, device_id: str, device_name: str, device_type: str) -> None:
        super().__init__(device_id, device_type, "BEGIN_PAIR")
        self.DEVICE_NAME = str(device_name)

    def process_response(self, json_obj: Dict[str, Any]) -> BeginPairResponse:
        item = get_json_obj(json_obj, ProtoConstants.RESPONSE_ITEM)
        response = BeginPairResponse(
            get_json_obj(item, ProtoConstants.CHALLENGE_TYPE),
            get_json_obj(item, ProtoConstants.PAIRING_REQ_TOKEN),
        )
        return response


class PairChallengeResponse(object):
    def __init__(self, auth_token: str) -> None:
        self.auth_token = auth_token


class PairChallengeCommand(PairCommandBase):
    """Finish pairing."""

    def __init__(
        self,
        device_id: str,
        challenge_type: int,
        pairing_token: int,
        pin: str,
        device_type: str,
    ) -> None:
        super().__init__(device_id, device_type, "FINISH_PAIR")
        self.CHALLENGE_TYPE = int(challenge_type)
        self.PAIRING_REQ_TOKEN = int(pairing_token)
        self.RESPONSE_VALUE = str(pin)

    def process_response(self, json_obj: Dict[str, Any]) -> PairChallengeResponse:
        item = get_json_obj(json_obj, ProtoConstants.RESPONSE_ITEM)
        response = PairChallengeResponse(get_json_obj(item, ProtoConstants.AUTH_TOKEN))
        return response


class CancelPairCommand(PairCommandBase):
    """Cancel pairing process."""

    def __init__(self, device_id, device_name: str, device_type: str) -> None:
        super().__init__(device_id, device_type, "CANCEL_PAIR")
        self.DEVICE_NAME = str(device_name)

    def process_response(self, json_obj: Dict[str, Any]) -> bool:
        return True



import asyncio
from functools import wraps
import logging
import sys

import click
from tabulate import tabulate

from . import VizioAsync
from .const import (
    DEFAULT_DEVICE_CLASS,
    DEVICE_CLASS_SOUNDBAR,
    DEVICE_CLASS_SPEAKER,
    DEVICE_CLASS_TV,
)

if sys.version_info < (3, 7):
    print("To use this script you need python 3.7 or newer, got %s" % sys.version_info)
    sys.exit(1)

_LOGGER = logging.getLogger(__name__)
DEVICE_ID = "pyvizio"
DEVICE_NAME = "Python Vizio"

pass_vizio = click.make_pass_decorator(VizioAsync)


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.group(invoke_without_command=False)
@click.option(
    "--ip",
    envvar="VIZIO_IP",
    required=True,
    help="IP of the device to connect to (optionally add custom port by specifying '<IP>:<PORT>')",
)
@click.option(
    "--auth",
    envvar="VIZIO_AUTH",
    required=False,
    help="Auth token for the device to connect to (refer to documentation on how to obtain auth token)",
)
@click.option(
    "--device_type",
    envvar="VIZIO_DEVICE_TYPE",
    required=False,
    default=DEFAULT_DEVICE_CLASS,
    type=click.Choice([DEVICE_CLASS_TV, DEVICE_CLASS_SPEAKER, DEVICE_CLASS_SOUNDBAR]),
)
@click.pass_context
def cli(ctx, ip: str, auth: str, device_type: str) -> None:
    logging.basicConfig(level=logging.INFO)
    ctx.obj = VizioAsync(DEVICE_ID, ip, DEVICE_NAME, auth, device_type)


@cli.command()
@click.option(
    "--timeout",
    required=False,
    default=10,
    type=click.IntRange(min=1),
    help="Number of seconds to wait for devices to be discovered",
)
def discover(timeout: int) -> None:
    logging.basicConfig(level=logging.INFO)
    devices = VizioAsync.discovery(timeout)
    table = tabulate(
        [[dev.ip, dev.port, dev.model, dev.name] for dev in devices],
        headers=["IP", "Port", "Model", "Name"],
    )
    _LOGGER.info("\n%s", table)


@cli.command()
@coro
@pass_vizio
async def pair(vizio: VizioAsync) -> None:
    _LOGGER.info(
        "Initiating pairing process, please check your device for pin upon success"
    )
    pair_data = await vizio.start_pair()
    if pair_data is not None:
        _LOGGER.info("Challenge type: %s", pair_data.ch_type)
        _LOGGER.info("Challenge token: %s", pair_data.token)


@cli.command()
@coro
@pass_vizio
async def pair_stop(vizio: VizioAsync) -> None:
    _LOGGER.info("Sending stop pair command")
    await vizio.stop_pair()


@cli.command()
@click.option(
    "--ch_type",
    required=False,
    default=1,
    type=click.INT,
    help="Challenge type obtained from pair command",
)
@click.option(
    "--token",
    required=True,
    type=click.INT,
    help="Challenge token obtained from pair command",
)
@click.option(
    "--pin",
    required=True,
    type=click.STRING,
    help="PIN obtained from device after running pair command",
)
@coro
@pass_vizio
async def pair_finish(vizio: VizioAsync, ch_type: int, token: int, pin: str) -> None:
    _LOGGER.info("Finishing pairing")
    pair_data = await vizio.pair(ch_type, token, pin)
    if pair_data is not None:
        _LOGGER.info("Authorization token: %s", pair_data.auth_token)


@cli.command()
@coro
@pass_vizio
async def input_list(vizio: VizioAsync) -> None:
    inputs = await vizio.get_inputs()
    if inputs:
        table = tabulate(
            [[input.name, input.meta_name, input.type, input.id] for input in inputs],
            headers=["Name", "Friendly Name", "Type", "ID"],
        )
        _LOGGER.info("\n%s", table)
    else:
        _LOGGER.error("Couldn't get available inputs")


@cli.command()
@coro
@pass_vizio
async def input_get(vizio: VizioAsync) -> None:
    data = await vizio.get_current_input()
    if data:
        _LOGGER.info("Current input: %s", data.meta_name)
    else:
        _LOGGER.error("Couldn't get current input")


@cli.command()
@coro
@pass_vizio
async def power_get(vizio: VizioAsync) -> None:
    is_on = await vizio.get_power_state()
    _LOGGER.info("Device is on" if is_on else "Device is off")


@cli.command()
@click.argument(
    "state",
    required=False,
    default="toggle",
    type=click.Choice(["toggle", "on", "off"]),
)
@coro
@pass_vizio
async def power(vizio: VizioAsync, state: str) -> None:
    if "on" == state:
        txt = "Turning ON"
        result = await vizio.pow_on()
    elif "off" == state:
        txt = "Turning OFF"
        result = await vizio.pow_off()
    else:
        txt = "Toggling power"
        result = await vizio.pow_toggle()
    _LOGGER.info(txt)
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument(
    "state", required=False, default="up", type=click.Choice(["up", "down"])
)
@click.argument(
    "amount", required=False, default=1, type=click.IntRange(1, 100, clamp=True)
)
@coro
@pass_vizio
async def volume(vizio: VizioAsync, state: str, amount: int) -> None:
    if "up" == state:
        txt = "Increasing volume"
        result = await vizio.vol_up(amount)
    else:
        txt = "Decreasing volume"
        result = await vizio.vol_down(amount)
    _LOGGER.info(txt)
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@coro
@pass_vizio
async def volume_current(vizio: VizioAsync) -> None:
    _LOGGER.info("Current volume: %s", await vizio.get_current_volume())


@cli.command()
@coro
@pass_vizio
async def volume_max(vizio: VizioAsync) -> None:
    _LOGGER.info("Max volume: %s", await vizio.get_max_volume())


@cli.command()
@click.argument(
    "state",
    required=False,
    default="previous",
    type=click.Choice(["up", "down", "previous"]),
)
@click.argument(
    "amount", required=False, default=1, type=click.IntRange(1, 100, clamp=True)
)
@coro
@pass_vizio
async def channel(vizio: VizioAsync, state: str, amount: str) -> None:
    amount = int(amount)
    if "up" == state:
        txt = "Channel up"
        result = await vizio.ch_up(amount)
    elif "down" == state:
        txt = "Channel down"
        result = await vizio.ch_down(amount)
    else:
        txt = "Previous channel"
        result = await vizio.ch_prev()
    _LOGGER.info(txt)
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument(
    "state",
    required=False,
    default="toggle",
    type=click.Choice(["toggle", "on", "off"]),
)
@coro
@pass_vizio
async def mute(vizio: VizioAsync, state: str) -> None:
    if "on" == state:
        txt = "Muting"
        result = await vizio.mute_on()
    elif "off" == state:
        txt = "Un-muting"
        result = await vizio.mute_off()
    else:
        txt = "Toggling mute"
        result = await vizio.mute_toggle()
    _LOGGER.info(txt)
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@coro
@pass_vizio
async def input_next(vizio: VizioAsync) -> None:
    result = await vizio.input_next()
    _LOGGER.info("Circling input")
    _LOGGER.info("OK" if result else "ERROR")


@cli.command(name="input")
@click.option(
    "--name", required=True, type=click.STRING, help="Input name to switch to"
)
@coro
@pass_vizio
async def input(vizio: VizioAsync, name: str) -> None:
    result = await vizio.input_switch(name)
    _LOGGER.info("Switching input")
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@coro
@pass_vizio
async def play(vizio: VizioAsync) -> None:
    result = await vizio.play()
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@coro
@pass_vizio
async def pause(vizio: VizioAsync) -> None:
    result = await vizio.pause()
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument("key", required=True, type=click.STRING)
@coro
@pass_vizio
async def key_press(vizio: VizioAsync, key: str) -> None:
    _LOGGER.info("Emulating pressing of '%s' key", key)
    result = await vizio.remote(key.upper())
    _LOGGER.info("OK" if result else "ERROR")


if __name__ == "__main__":
    cli()



from pyvizio.version import __version__
from setuptools import setup

with open("README.md", "r") as myfile:
    longdescription = myfile.read()

setup(
    name="pyvizio",
    version=__version__,
    description="Python library for interfacing with Vizio SmartCast TVs and Sound Bars (2016+ models)",
    long_description=longdescription,
    long_description_content_type="text/markdown",
    url="https://github.com/vkorn/pyvizio",
    author="Vlad Korniev",
    author_email="vladimir.kornev@gmail.com",
    license="GPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
    ],
    keywords="vizio smartcast",
    packages=["pyvizio"],
    install_requires=[
        "aiohttp",
        "asyncio",
        "click",
        "jsonpickle",
        "tabulate",
        "zeroconf",
    ],
    entry_points={"console_scripts": ["pyvizio=pyvizio.cli:cli"]},
)
